"""
Serial implementation of k-means clustering in PyTorch
"""
import time

import h5py
import torch


class KMeans:
    """
    Serial k-means clustering in PyTorch.

    Attributes
    ----------
    n_clusters : int
        The number of clusters, i.e., k.
    max_iter : int
        The maximum number of iterations to perform.
    tol : float
        The tolerance for the convergence criterion.
    _centroids : Union[None, torch.Tensor]
        The current centroids.
    _matching_centroids : Union[None, torch.Tensor]
        Assigned centroids for all samples in dataset.
    _inertia : float
        The inertia (quantity to be checked for convergence).

    Methods
    -------
    _initialize_centroids(x)
        Randomly initialize centroids.
    _fit_to_cluster(x)
        Get the closest centroid for each sample in dataset.
    fit(x)
        Perform k-means clustering.
    """

    def __init__(
        self, n_clusters: int = 8, max_iter: int = 300, tol: float = -1.0
    ) -> None:
        """
        Configure k-means clustering algorithm.

        Parameters
        ----------
        n_clusters : int
            The number of clusters, i.e., k.
        max_iter : int
            The maximum number of iterations to be performed.
        tol : float
            The tolerance for the convergence criterion.
        """
        self.n_clusters = n_clusters  # Number of clusters
        self.max_iter = max_iter  # Maximum number of iterations
        self._centroids = None
        self._matching_centroids = None
        self.tol = tol  # Tolerance for convergence criterion
        self._inertia = float("nan")

    def _initialize_centroids(self, x: torch.Tensor) -> None:
        """
        Randomly initialize the centroids.

        Parameters
        ----------
        x : torch.Tensor
            The dataset to be clustered.
        """
        # Shuffle data and choose first `n_clusters` samples as initial centroids.
        self._centroids = x[torch.randperm(x.shape[0])[: self.n_clusters]]

    def _fit_to_cluster(self, x: torch.Tensor) -> torch.Tensor:
        """
        Determine the closest centroids for each sample in dataset as measured by their Euclidean distance.

        Parameters
        ----------
        x : torch.Tensor
            The dataset to be clustered.

        Returns
        -------
        torch.Tensor
            Indices of matching centroids for each sample in dataset.
        """
        distances = torch.cdist(
            x, self._centroids
        )  # Calculate Euclidean distance of each data sample to each current centroid.
        return distances.argmin(
            dim=1, keepdim=True
        )  # Return index of the closest centroid for each sample.

    def fit(self, x: torch.Tensor) -> "KMeans":
        """
        Perform k-means clustering of given dataset.

        Parameters
        ----------
        x : torch.Tensor
            The dataset to cluster.

        Returns
        -------
        KMeans
            The fitted KMeans object containing final centroids.
        """
        self._initialize_centroids(x)  # Initialize centroids.
        new_cluster_centers = self._centroids.clone()

        # Iteratively fit points to centroids.
        for idx in range(self.max_iter):
            # Determine index of the closest centroid for each sample in dataset.
            print(f"Iteration {idx}...")
            self._matching_centroids = self._fit_to_cluster(
                x
            )  # Array of length `n_samples` providing index of closest centroid for each sample in dataset.

            # Update centroids.
            for i in range(self.n_clusters):  # Loop over clusters.
                # Determine all points in current cluster.
                selection_mask = (self._matching_centroids == i).type(torch.int64)
                # Array of length `n_samples` with binary encoding of whether each sample belongs to cluster i or not.

                assigned_points = (x * selection_mask).sum(
                    axis=0, keepdim=True
                )  # Compute vectorial sum of all points in current cluster.
                points_in_cluster = selection_mask.sum(axis=0, keepdim=True).clamp(
                    1, torch.iinfo(torch.int64).max
                )  # Compute number of points in current cluster.
                new_cluster_centers[i : i + 1, :] = (
                    assigned_points / points_in_cluster
                )  # Compute new centroids.

            # Check whether centroid movement has converged.
            self._inertia = (
                (self._centroids - new_cluster_centers) ** 2
            ).sum()  # Update inertia.
            self._centroids = new_cluster_centers.clone()
            if (
                self.tol is not None and self._inertia <= self.tol
            ):  # Check whether inertia is smaller than tolerance.
                break

        return self


if __name__ == "__main__":
    print(
        "##############################\n"
        "# PyTorch k-Means Clustering #\n"
        "##############################"
    )

    data_path = "/pfs/work7/workspace/scratch/ku4408-VL_ScalableAI/data/cityscapes_300.h5"
    dataset = "cityscapes_data"
    device = torch.device(
        "cuda" if torch.cuda.is_available() else "cpu"
    )  # Set device as available.

    if torch.cuda.is_available():
        print("Running on GPU...")
    else:
        print("Running on CPU...")

    print(f"Loading dataset from {data_path}[{dataset}]...")
    # Data is available in HDF5 format.
    # An HDF5 file is a container for two kinds of objects:
    # - datasets: array-like collections of data
    # - groups: folder-like containers holding datasets and other groups
    # Most fundamental thing to remember when using h5py is:
    # Groups work like dictionaries, and datasets work like NumPy arrays.

    # Open file for reading. We use the Cityscapes dataset.
    with h5py.File(data_path, "r") as handle:
        print("Open h5 file...")
        data = torch.tensor(
            handle[dataset][:300], device=device
        )  # Default device is "cpu", set to "cuda" for GPU.
    print("Torch tensor created.")

    # k-means hyperparameters
    num_clusters = 8
    num_iterations = 20

    kmeans_clusterer = KMeans(n_clusters=num_clusters, max_iter=num_iterations)
    print("Start fitting the data...")
    start = time.perf_counter()  # Start timer.
    kmeans_clusterer.fit(data)  # Perform k-means clustering.
    print(
        f"DONE.\nRun time: \t{time.perf_counter() - start} s"
    )  # Print measured runtime.
