"""
Feature-parallel implementation of k-means clustering in PyTorch using MPI
"""
import time

import h5py
import torch
from mpi4py import MPI


class KMeans:
    """
    Feature-parallel k-means clustering in PyTorch with MPI.

    Each rank holds an exclusive chunk of the complete dataset (distributed along the feature axis, i.e., each rank
    holds all samples but only an exclusive fraction of their features). and determines
    locally which local sample belongs to which cluster before updating the centroids globally based on these local
    assignments and continuing with the next iteration.

    Feature parallelism for k-means requires communication at two points in the code, i.e.:
    1. Centroid initialization: Either, one rank determines the initial centroids and broadcasts them to all other ranks
       or all ranks randomly choose the initial centroids using the same seed.
    2. Cluster assignment: To determine the nearest centroid for each sample, we need to calculate its distance to each
       cluster center considering all features. To achieve this, we compute the contribution of each local fraction of
       features to the squared distance before summing these local squared distances up to obtain a global value.
    3. Convergence check: Again, the local contributions to the inertia need to be summed up to obtain a global value.


    Attributes
    ----------
    comm : MPI.Comm
        The MPI communicator.
    n_clusters : int
        The number of clusters.
    max_iter : int
        The maximum number of iterations.
    tol : float
        The convergence criterion.
    _inertia : float
        The inertia (quantity to be checked for convergence).
    _cluster_centers : Union[None, torch.Tensor]
        The centroids.
    _matching_centroids : torch.Tensor
        Indices of nearest centroid for each sample.
    """

    def __init__(
        self,
        comm: MPI.Comm = MPI.COMM_WORLD,
        n_clusters: int = 8,
        max_iter: int = 300,
        tol: float = -1.0,
    ) -> None:
        """
        Configure feature-parallel k-means clustering algorithm.

        Parameters
        ----------
        comm : MPI.Comm
            The MPI communicator.
        n_clusters : int
            The number of clusters, i.e., k.
        max_iter : int
            The maximum number of iterations to be performed.
        tol : float
            The tolerance for the convergence criterion.
        """
        self.comm = comm  # MPI communicator
        self.n_clusters = n_clusters  # Number of clusters
        self.max_iter = max_iter  # Maximum number of iterations
        self.tol = tol  # Tolerance for convergence criterion

        self._inertia = float("nan")
        self._cluster_centers = None
        self._matching_centroids = None

    def _initialize_centroids(self, x: torch.Tensor) -> None:
        """
        Randomly initialize the centroids.

        For feature parallelism, all samples of the dataset are partially present on each rank.

        Parameters
        ----------
        x : torch.Tensor
            The dataset to be clustered.
        """
        print(
            f"Rank {self.comm.rank}/{self.comm.size}: Shape of data [#samples, #features] = {list(x.shape)}"
        )
        # As all samples are partially present on each rank, only shuffle on root to determine centroids.
        # Alternatively, all ranks can shuffle using the same seed.
        if rank == 0:
            indices = torch.randperm(x.shape[0])[
                : self.n_clusters
            ]  # Choose first `n_cluster` randomly shuffled indices as centroid indices.

        # `torch.randperm(n)` returns random permutation of integers from 0 to n - 1.
        # `x.shape[0]` gives number of samples in data.
        else:
            indices = torch.empty(self.n_clusters, dtype=torch.int64)
        self.comm.Bcast(indices, root=0)
        self._cluster_centers = x[indices]  # Set centroids as class attribute.

    def _fit_to_cluster(self, x: torch.Tensor) -> torch.Tensor:
        """
        Determine the closest centroids for each sample in dataset as measured by their Euclidean distance.

        Parameters
        ----------
        x : torch.Tensor
            The dataset to be clustered.

        Returns
        -------
        torch.Tensor
            Indices of matching centroids for each sample in dataset.
        """
        squared_distances_local = (
            torch.cdist(x, self._cluster_centers) ** 2
        )  # Calculate squared distance between samples and centroids for locally present fraction of features.
        squared_distances_global = torch.empty(
            squared_distances_local.shape, dtype=torch.float
        )  # Initialize variable for global squared distance between samples and centroids for all features.
        self.comm.Allreduce(
            squared_distances_local, squared_distances_global, op=MPI.SUM
        )  # Sum up squared distances over features.
        return squared_distances_global.argmin(
            dim=1, keepdim=True
        )  # Return index of nearest centroid for each sample.

    def fit(self, x: torch.Tensor) -> "KMeans":
        """
        Perform k-means clustering of given dataset.

        Parameters
        ----------
        x : torch.Tensor
            The dataset to cluster.

        Returns
        -------
        KMeans
            The fitted KMeans object containing final centroids.
        """
        self._initialize_centroids(x)  # Initialize centroids.
        new_cluster_centers = self._cluster_centers.clone()

        # Iteratively fit points to centroids.
        for idx in range(self.max_iter):
            # Determine nearest centroids.
            print(f"Rank {self.comm.rank}/{self.comm.size}: Iteration {idx}...")
            self._matching_centroids = self._fit_to_cluster(
                x
            )  # Redundantly determine index of nearest centroid for each sample.

            # Update centroids.
            for i in range(self.n_clusters):
                # Locally determine samples in currently considered cluster i (binary encoding).
                selection_mask = (self._matching_centroids == i).type(torch.int64)

                # Locally accumulate samples and total number of samples in cluster i.
                assigned_points = (x * selection_mask).sum(
                    axis=0, keepdim=True
                )  # Sum up samples in cluster i.
                points_in_cluster = selection_mask.sum(
                    axis=0, keepdim=True
                ).clamp(  # Determine overall number of samples in cluster i.
                    1.0, torch.iinfo(torch.int64).max
                )

                # Compute new centroids.
                new_cluster_centers[
                    i : i + 1, :
                ] = assigned_points / points_in_cluster.clamp(
                    1.0, torch.iinfo(torch.int64).max
                )

            # Check whether centroid movement has converged.
            inertia_local = ((self._cluster_centers - new_cluster_centers) ** 2).sum()
            inertia_global = torch.empty(inertia_local.shape)
            self.comm.Allreduce(inertia_local, inertia_global, op=MPI.SUM)
            self._inertia = inertia_global
            self._cluster_centers = new_cluster_centers.clone()
            if self.tol is not None and self._inertia <= self.tol:
                break
        return self


if __name__ == "__main__":
    comm = MPI.COMM_WORLD
    rank = comm.rank
    size = comm.size

    if rank == 0:
        print(
            "###############################################\n"
            "# Feature-parallel PyTorch k-means clustering #\n"
            "###############################################"
        )

    data_path = "/pfs/work7/workspace/scratch/ku4408-VL_ScalableAI/data/cityscapes_300.h5"
    dataset = "cityscapes_data"

    if rank == 0:
        print(f"\nLoading data... {data_path}[{dataset}]\n")

    # Data is available in HDF5 format.
    # An HDF5 file is a container for two kinds of objects:
    # - datasets: array-like collections of data
    # - groups: folder-like containers holding datasets and other groups
    # Most fundamental thing to remember when using h5py is:
    # Groups work like dictionaries, and datasets work like NumPy arrays.
    with h5py.File(data_path, "r") as handle:
        chunk = int(
            handle[dataset].shape[1] / size
        )  # Calculate local number of features in each chunk.
        if (
            rank == size - 1
        ):  # Load data chunks from file (distributed along feature axis).
            data = torch.tensor(handle[dataset][:, rank * chunk :])
        else:
            data = torch.tensor(handle[dataset][:, rank * chunk : (rank + 1) * chunk])

    print(f"Rank {rank}/{size}: \t[OK]")

    # k-means hyperparameters
    num_clusters = 8
    num_iterations = 20

    kmeans_clusterer = KMeans(
        comm=comm, n_clusters=num_clusters, max_iter=num_iterations
    )
    if rank == 0:
        print("Start fitting the data...")
        start = time.perf_counter()  # Start runtime measurement.

    kmeans_clusterer.fit(data)  # Perform actual k-means clustering.

    if rank == 0:
        print(
            f"DONE.\nRun time:\t{time.perf_counter() - start} s"
        )  # Print measured runtime.
